from hledger_lots import fifo

from . import lots_data


class TestGetLots:
    def test_only_buying(self):
        assert (
            fifo.get_lots(lots_data.txns_only_buying, check=False)
            == lots_data.txns_only_buying
        )

    def test_never_zero(self):
        assert (
            fifo.get_lots(lots_data.txns_qtty_never_zero, check=False)
            == lots_data.expected_qtty_never_zero
        )

    def test_qtty_reach_zero(self):
        assert (
            fifo.get_lots(lots_data.txns_qtty_reaches_zero, check=False)
            == lots_data.expected_qtty_reaches_zero
        )


class TestGetSellLots:
    def test_sell_all(self):
        sell_lots = fifo.get_sell_lots(
            lots_data.txns_qtty_reaches_zero,
            sell_date="2022-02-01",
            sell_qtty=5,
            check=False,
        )
        assert sell_lots == lots_data.expected_qtty_reaches_zero_sell_all

    def test_sell_some(self):
        sell_lots = fifo.get_sell_lots(
            lots_data.txns_qtty_never_zero,
            sell_date="2022-02-01",
            sell_qtty=11,
            check=False,
        )
        assert sell_lots == lots_data.expected_qtty_never_zero_sell_some


class TestTxn2Hl:
    txns = lots_data.expected_qtty_reaches_zero_sell_all
    date = "2022-02-01"
    cash_account = "Bank"
    revenue_account = "Revenue"

    def test_txn2hl_profit(self):
        cur = "USD"
        test = fifo.txn2hl(
            self.txns, self.date, cur, self.cash_account, self.revenue_account, 160, ""
        )

        expected = """2022-02-01 Sold USD  ; cost_method:fifo
    ; commodity:USD, qtty:5.00, price:32.00
    ; avg_cost:26.0000, xirr:61.42% annual percent rate 30/360US
    ; command:
    Bank                 160.00 USD
    Acct1      -2.00 USD @ 35.0 USD  ; buy_date:2022-01-12, base_cur:USD
    Acct1      -3.00 USD @ 20.0 USD  ; buy_date:2022-01-14, base_cur:USD
    Revenue              -30.00 USD

"""

        assert test == expected

    def test_txn2hl_loss(self):
        cur = "USD"
        test = fifo.txn2hl(
            self.txns, self.date, cur, self.cash_account, self.revenue_account, 80, ""
        )

        expected = """2022-02-01 Sold USD  ; cost_method:fifo
    ; commodity:USD, qtty:5.00, price:16.00
    ; avg_cost:26.0000, xirr:-1.00% annual percent rate 30/360US
    ; command:
    Bank                  80.00 USD
    Acct1      -2.00 USD @ 35.0 USD  ; buy_date:2022-01-12, base_cur:USD
    Acct1      -3.00 USD @ 20.0 USD  ; buy_date:2022-01-14, base_cur:USD
    Revenue               50.00 USD

"""

        assert test == expected



from pathlib import Path

import pytest

from hledger_lots.lib import AdjustedTxn, Txn
from hledger_lots.hl import adjust_txn, hledger2txn

from . import lots_data


class TestAdjustTxn:
    def test_adjust_txn_unit_price(self):
        txn = Txn(
            date="2022-01-01",
            price=10.0,
            base_cur="USD",
            qtty=1.0,
            acct="Acct1",
            type="UnitPrice",
        )
        expected = AdjustedTxn(
            date="2022-01-01", price=10.0, base_cur="USD", qtty=1.0, acct="Acct1"
        )
        assert adjust_txn(txn) == expected

    def test_adjust_txn_total_price(self):
        txn = Txn(
            date="2022-01-01",
            price=20.0,
            base_cur="USD",
            qtty=2.0,
            acct="Acct1",
            type="TotalPrice",
        )
        expected = AdjustedTxn(
            date="2022-01-01", price=10.0, base_cur="USD", qtty=2.0, acct="Acct1"
        )
        assert adjust_txn(txn) == expected

    def test_adjust_txn_zero_quantity(self):
        txn = Txn(
            date="2022-01-01",
            price=10.0,
            base_cur="USD",
            qtty=0.0,
            acct="Acct1",
            type="TotalPrice",
        )
        with pytest.raises(ZeroDivisionError):
            adjust_txn(txn)


class TestHledger2Txn:
    hl_txns = """2023-01-05 Buy AAPL
    Asset:Stocks                                   3 AAPL @ 5.2 USD
    Asset:Bank

2023-01-10 Buy BRL
    Asset:FOREX                                  55 BRL @@ 10 USD
    Asset:Bank

2023-01-05 Sell AAPL
    Asset:Bank                                    20 BRL
    Asset:Stocks                                   3 AAPL @ 5.2 USD
    Revenue:Capital Gain
"""

    def test_price_qtty(self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
        expected = [
            AdjustedTxn(
                date="2023-01-05",
                price=5.2,
                base_cur="USD",
                qtty=3,
                acct="Asset:Stocks",
            ),
            AdjustedTxn(
                date="2023-01-05",
                price=5.2,
                base_cur="USD",
                qtty=3,
                acct="Asset:Stocks",
            ),
        ]

        monkeypatch.setattr("sys.stdin", None)

        file_path = tmp_path.joinpath("data.journal")
        file_path.touch()
        file_path.write_text(self.hl_txns)
        file_tup = (str(file_path),)

        assert hledger2txn(file_tup, "AAPL") == expected

    def test_total_amount(self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
        expected = [
            AdjustedTxn(
                date="2023-01-10",
                price= 10/55,
                base_cur="USD",
                qtty=55,
                acct="Asset:FOREX",
            )
        ]

        monkeypatch.setattr("sys.stdin", None)

        file_path = tmp_path.joinpath("data.journal")
        file_path.touch()
        file_path.write_text(self.hl_txns)
        file_tup = (str(file_path),)

        assert hledger2txn(file_tup, "BRL") == expected



from typing import List

from .lib import AdjustedTxn


class MultipleBaseCurrencies(Exception):
    def __init__(self, currencies: set) -> None:
        self.currencies = currencies
        self.message = f"More than one base currency: {currencies}"


def check_short_sell_past(previous_buys: List[AdjustedTxn], sell: AdjustedTxn):
    previous_buys_qtty = sum([txn.qtty for txn in previous_buys])
    if abs(sell.qtty) > abs(previous_buys_qtty):
        raise ValueError(f"Short sell not allowed for sell {sell}")


def check_short_sell_current(previous_buys: List[AdjustedTxn], sell_qtty: float):
    previous_buys_qtty = sum([txn.qtty for txn in previous_buys])
    if sell_qtty > previous_buys_qtty:
        raise ValueError(
            f"Short sell not allowed for sell. You have {previous_buys_qtty:.4f}, which is less than you want to sell: {sell_qtty:.4f}"
        )


def check_base_currency(txns: List[AdjustedTxn]):
    base_currencies = set(txn.base_cur for txn in txns)
    if len(base_currencies) > 1:
        raise MultipleBaseCurrencies(base_currencies)


def check_available(txns: List[AdjustedTxn], account: str, sell_qtty: float):
    sum_qtty = sum(txn.qtty for txn in txns if txn.acct == account)
    if sell_qtty > sum_qtty:
        raise ValueError(
            f"You can't sell {sell_qtty:,.2f} in {account} because there is only {sum_qtty:,.2f} available"
        )



import os
from pathlib import Path
from typing import List, Optional, Tuple

import click

ENV_FILE = "LEDGER_FILE"
default_path = Path.home() / ".hledger.journal"


def get_default_file() -> Tuple[str, ...]:
    file_env = os.getenv("LEDGER_FILE")
    if file_env:
        return (file_env,)
    elif default_path.exists():
        return (str(default_path),)
    else:
        raise click.BadOptionUsage("file", "File missing")


def get_file_path(
    ctx: click.Context, _param, value: Optional[Tuple[str, ...]]
) -> Optional[Tuple[str, ...]]:  # type: ignore
    if value:
        return value

    if not ctx.parent:
        return None

    filenames: Optional[Tuple] = ctx.parent.params.get("file")
    if not filenames:
        raise click.BadOptionUsage("file", "File missing")

    return filenames


def get_files_comm(file_path: Tuple[str, ...]) -> List[str]:
    files = []
    for file in file_path:
        files = [*files, "-f", file]
    return files



import json
import subprocess
import sys
from typing import List, Optional, Tuple

from .files import get_files_comm
from .lib import AdjustedTxn, Txn, get_avg_fifo, get_xirr


def adjust_txn(txn: Txn) -> AdjustedTxn:
    price = txn.price if txn.type == "UnitPrice" else txn.price / txn.qtty

    result = AdjustedTxn(txn.date, price, txn.base_cur, txn.qtty, txn.acct)
    return result


def prices_items2txn(date: str, prices_items: dict, account: str) -> Txn:
    price = prices_items["aprice"]["contents"]["aquantity"]["floatingPoint"]
    base_cur = prices_items["aprice"]["contents"]["acommodity"]
    qtty = prices_items["aquantity"]["floatingPoint"]
    price_type = prices_items["aprice"]["tag"]

    txn = Txn(date, price, base_cur, qtty, account, price_type)
    return txn


def hledger2txn(
    file_path: Tuple[str, ...], cur: str, no_desc: Optional[str] = None
) -> List[AdjustedTxn]:
    files_comm = get_files_comm(file_path)
    comm = ["hledger", *files_comm, "print", f"cur:{cur}", "--output-format=json"]
    if no_desc:
        comm.append(f"not:desc:{no_desc}")

    hl_proc = subprocess.run(comm, stdin=sys.stdin, capture_output=True)
    if hl_proc.returncode != 0:
        raise ValueError(hl_proc.stderr.decode("utf8"))

    hl_data = hl_proc.stdout.decode("utf8")
    txns_list = json.loads(hl_data)

    txns = [
        prices_items2txn(txn["tdate"], prices_items, posting_items["paccount"])
        for txn in txns_list
        for posting_items in txn["tpostings"]
        for prices_items in posting_items["pamount"]
        if prices_items["acommodity"] == cur and prices_items["aprice"]
    ]

    adjusted_txns = [adjust_txn(txn) for txn in txns]
    return adjusted_txns



import pytest
from hledger_lots import files
import click


class PathMissing:
    @staticmethod
    def exists():
        return False


class TestGetDefaultFile:
    def test_with_file_env(self, monkeypatch: pytest.MonkeyPatch):
        file_env = "/path/to/file.dat"
        monkeypatch.setenv("LEDGER_FILE", file_env)
        assert files.get_default_file() == (file_env,)

    def test_with_default_path(self, monkeypatch, tmp_path):
        default_path = tmp_path.home() / ".hledger.journal"
        monkeypatch.setattr("hledger_lots.files.default_path", default_path)
        assert files.get_default_file() == (str(default_path),)

    def test_missing_file(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.delenv("LEDGER_FILE", raising=False)
        monkeypatch.setattr("hledger_lots.files.default_path", PathMissing())

        with pytest.raises(click.BadOptionUsage):
            files.get_default_file()


class TestGetFilePath:
    def test_has_value(self):
        value = ("a.journal", "b.journal")
        context = click.Context(click.Command("command"))
        assert files.get_file_path(context, None, value) == value

    def test_no_parent(self):
        context = click.Context(click.Command("command"))
        assert files.get_file_path(context, None, None) is None

    def test_get_parent(self):
        default_file = ("default.journal",)
        parent = click.Context(click.Command("parent"))
        parent.params = dict(file=default_file)
        context = click.Context(click.Command("command"), parent=parent)
        assert files.get_file_path(context, None, None) == default_file


class TestGetFilesComm:
    def test_empty_path(self):
        assert files.get_files_comm(tuple()) == []

    def test_single_file(self):
        assert files.get_files_comm(("file1.txt",)) == ["-f", "file1.txt"]

    def test_multiple_files(self):
        assert files.get_files_comm(("file1.txt", "file2.txt", "file3.txt")) == [
            "-f",
            "file1.txt",
            "-f",
            "file2.txt",
            "-f",
            "file3.txt",
        ]

    def test_duplicate_files(self):
        assert files.get_files_comm(("file1.txt", "file1.txt")) == [
            "-f",
            "file1.txt",
            "-f",
            "file1.txt",
        ]



import os
import re
import shlex
from dataclasses import asdict, dataclass
from datetime import date
from typing import List, Optional

from pyxirr import DayCount, xirr
from tabulate import tabulate

from .files import get_files_comm


@dataclass
class AdjustedTxn:
    date: str
    price: float
    base_cur: str
    qtty: float
    acct: str


@dataclass
class Txn(AdjustedTxn):
    type: str


class CostMethodError(Exception):
    def __init__(self, sell: AdjustedTxn, price: float, base_cur: str) -> None:
        self.message = f"Error in sale {sell}. Correct price should be {price} in currency {base_cur}"
        super().__init__(self.message)


def get_avg_fifo(txns: List[AdjustedTxn]):
    total_qtty = sum(txn.qtty for txn in txns)
    if total_qtty == 0:
        return 0
    mult = [txn.qtty * txn.price for txn in txns]
    total_mult = sum(mult)
    avg = total_mult / total_qtty
    return avg


def get_xirr(
    sell_price: float, sell_date: date, txns: List[AdjustedTxn]
) -> Optional[float]:
    if len(txns) == 0:
        return 0

    dates = [txn.date for txn in txns]
    buy_amts = [txn.price * txn.qtty for txn in txns]
    total_qtty = sum(txn.qtty for txn in txns)

    sell_date_txt = sell_date.strftime("%Y-%m-%d")
    dates = [*dates, sell_date_txt]
    amts = [*buy_amts, -total_qtty * sell_price]
    sell_xirr = xirr(dates, amts, day_count=DayCount.THIRTY_U_360)
    return sell_xirr


def dt_list2table(dt_list: List, tablefmt: str = "simple"):
    lots_dict = [asdict(dt) for dt in dt_list]
    table = tabulate(
        lots_dict,
        headers="keys",
        numalign="decimal",
        floatfmt=",.4f",
        tablefmt=tablefmt,
    )
    return table


def adjust_commodity(comm: str):
    has_non_word = re.search(r"\W", comm)
    adjusted = f'"{comm}"' if has_non_word else comm
    return adjusted


def default_fn_bool(env_name: str, default: bool):
    env = os.environ.get(env_name, None)
    if env == "true":
        return True
    elif env == "False":
        return False
    else:
        return default


def get_sell_comm(
    commodity: str,
    no_desc: str,
    commodity_account: str,
    cash_account: str,
    revenue_account: str,
    date: str,
    quantity: float,
    price: float,
    avg_cost: bool,
):
    avg_comm = ["-g"] if avg_cost else []
    no_desc_comm = ["n", no_desc] if no_desc else []

    comm = [
        "hledger-lots",
        "sell",
        *avg_comm,
        *no_desc_comm,
        "-c",
        commodity,
        "-s",
        commodity_account,
        "-a",
        cash_account,
        "-r",
        revenue_account,
        "-d",
        date,
        "-q",
        str(quantity),
        "-p",
        str(price),
    ]
    comm_str: str = shlex.join(comm)

    return comm_str



from hledger_lots.avg import get_avg_cost, avg_sell, AvgCost
from datetime import date
from . import lots_data


class TestGetAvgCost:
    def test_empty(self):
        assert get_avg_cost([], False, date(2022, 1, 1)) == []

    def test_only_buying_after_some(self):
        txns = lots_data.txns_only_buying
        expected = [
            AvgCost(
                date="2022-01-01", total_qtty=10.0, total_amount=100.0, avg_cost=10.0
            ),
            AvgCost(
                date="2022-01-02",
                total_qtty=12.0,
                total_amount=140.0,
                avg_cost=11.666666666666666,
            ),
            AvgCost(
                date="2022-01-03",
                total_qtty=17.0,
                total_amount=290.0,
                avg_cost=17.058823529411764,
            ),
            AvgCost(
                date="2022-01-04",
                total_qtty=65.0,
                total_amount=1010.0,
                avg_cost=15.538461538461538,
            ),
        ]

        assert get_avg_cost(txns, False,date(2022, 1, 4)) == expected

    def test_only_buying_after_all(self):
        txns = lots_data.txns_only_buying
        expected = [
            AvgCost(
                date="2022-01-01", total_qtty=10.0, total_amount=100.0, avg_cost=10.0
            ),
            AvgCost(
                date="2022-01-02",
                total_qtty=12.0,
                total_amount=140.0,
                avg_cost=11.666666666666666,
            ),
            AvgCost(
                date="2022-01-03",
                total_qtty=17.0,
                total_amount=290.0,
                avg_cost=17.058823529411764,
            ),
            AvgCost(
                date="2022-01-04",
                total_qtty=65.0,
                total_amount=1010.0,
                avg_cost=15.538461538461538,
            ),
            AvgCost(
                date="2022-01-05",
                total_qtty=86.0,
                total_amount=1535.0,
                avg_cost=17.848837209302324,
            ),
        ]

        assert get_avg_cost(txns, False) == expected

    def test_with_sell(self):
        txns = lots_data.txns_qtty_never_zero
        expected = [
            AvgCost(
                date="2022-01-01", total_qtty=5.0, total_amount=50.0, avg_cost=10.0
            ),
            AvgCost(
                date="2022-01-02",
                total_qtty=7.0,
                total_amount=90.0,
                avg_cost=12.857142857142858,
            ),
            AvgCost(
                date="2022-01-03",
                total_qtty=14.0,
                total_amount=300.0,
                avg_cost=21.428571428571427,
            ),
            AvgCost(
                date="2022-01-04",
                total_qtty=11.0,
                total_amount=235.71428571428572,
                avg_cost=21.42857142857143,
            ),
            AvgCost(
                date="2022-01-05",
                total_qtty=13.0,
                total_amount=285.7142857142857,
                avg_cost=21.978021978021978,
            ),
            AvgCost(
                date="2022-01-06",
                total_qtty=18.0,
                total_amount=460.7142857142857,
                avg_cost=25.595238095238095,
            ),
            AvgCost(
                date="2022-01-07",
                total_qtty=13.0,
                total_amount=332.73809523809524,
                avg_cost=25.595238095238095,
            ),
            AvgCost(
                date="2022-01-08",
                total_qtty=15.0,
                total_amount=372.73809523809524,
                avg_cost=24.849206349206348,
            ),
        ]

        assert get_avg_cost(txns, False,date(2022, 1, 8)) == expected


class TestAvgSell:
    def test_avg_sell(self):
        txns = lots_data.txns_only_buying
        test = avg_sell(
            txns=txns,
            date="2022-02-01",
            qtty=50,
            cur="AAPL",
            cash_account="Asset:Bank",
            revenue_account="Revenue:Capital Gain",
            comm_account="Acct1",
            value=1000,
            check=False,
            sell_cmd=""
        )

        expected = """2022-02-01 Sold AAPL  ; cost_method:avg_cost
    ; commodity:AAPL, qtty:50.00, price:20.00
    ; xirr:3.56% annual percent rate 30/360US
    ; command:
    Asset:Bank                                    1000.00 USD
    Acct1                   -50 AAPL @ 17.848837209302324 USD
    Revenue:Capital Gain               -107.5581395348838 USD

"""

        assert test == expected



import re
import sys
from dataclasses import dataclass
from datetime import date, datetime, timedelta
from pathlib import Path
from typing import List, Optional, Tuple

import yfinance as yf
from requests.exceptions import HTTPError
from requests_cache import CachedSession

from .hl import hledger2txn
from .info import get_commodities, get_last_price
from .lib import adjust_commodity, get_files_comm


@dataclass
class Price:
    name: str
    date: date
    price: float
    cur: str


def get_start_date(txn_first_date: date, last_market_date: Optional[date]):
    if not last_market_date:
        last_date = txn_first_date
    elif last_market_date < txn_first_date:
        last_date = txn_first_date
    else:
        last_date = last_market_date

    start_date = last_date + timedelta(days=1)
    start_date_str = start_date.strftime("%Y-%m-%d")
    return start_date_str


def filter_yahoo(com: List[str]):
    y_match = (re.search(r"^(y\.)(.*)", item) for item in com)
    y_match = (item for item in y_match if item)
    y_match = [match.groups()[1] for match in y_match]
    return y_match


def prices2hledger(prices: List[Price]):
    prices_list = [
        f"P {price.date.strftime('%Y-%m-%d')} \"y.{price.name}\" {price.price} {price.cur}"
        for price in prices
        if price
    ]
    prices_str = "\n".join(prices_list)
    return prices_str


def get_yahoo_prices(
    ticker_name: str,
    start_date: str,
    end_date: str,
    session: CachedSession,
):
    ticker = yf.Ticker(ticker_name, session=session)
    info = ticker.info

    if start_date:
        df = ticker.history(start=start_date, end=end_date, raise_errors=True)
    else:
        df = ticker.history(period="1d", raise_errors=True)

    prices = [
        Price(
            ticker_name,
            row[0].to_pydatetime().date(),  # type:ignore
            row[1]["Close"],  # type: ignore
            info["currency"],
        )
        for row in df.iterrows()
    ]
    return prices


def get_hledger_prices(files: Tuple[str, ...], append_prices_to: Path):
    files_comm = get_files_comm(files)
    commodities = get_commodities(files)
    tickers = filter_yahoo(commodities)
    tickers_str = " ".join(tickers)
    print(
        f"stderr: Downloading price history for tickers: {tickers_str}", file=sys.stderr
    )

    session_path = Path.home() / "yfinance.cache"
    session = CachedSession(str(session_path))
    today = datetime.today()
    yesterday = today - timedelta(days=1)
    yesterday_str = yesterday.strftime("%Y-%m-%d")
    with open(append_prices_to, "a") as f:
        f.write("\n")
        for ticker in tickers:
            commodity = f"y.{ticker}"
            txns = hledger2txn(files, commodity)
            txn_first_date_str = txns[0].date
            txn_first_date = datetime.strptime(txn_first_date_str, "%Y-%m-%d").date()

            last_market_date = get_last_price(files_comm, commodity)[0]
            start_date = get_start_date(txn_first_date, last_market_date)
            days_past = today.date() - datetime.strptime(start_date, "%Y-%m-%d").date()

            if days_past.days > 0:
                try:
                    prices = get_yahoo_prices(
                        ticker, start_date, yesterday_str, session
                    )
                    prices_hledger = prices2hledger(prices)
                    f.write(prices_hledger + "\n")
                except HTTPError:
                    print(f"stderr: {ticker} not found", file=sys.stderr)
                except Exception as e:
                    print(
                        f"stderr: Nothing downloaded for {ticker} between {start_date} and {yesterday_str}",
                        file=sys.stderr,
                    )

        f.write("\n")



import copy
import subprocess
from datetime import datetime
from typing import List

from . import checks
from .lib import AdjustedTxn, CostMethodError, adjust_commodity, get_avg_fifo, get_xirr


def check_sell(sell: AdjustedTxn, previous_buys: List[AdjustedTxn], check: bool):
    if not check:
        return

    diff_zero = [
        previous_buy for previous_buy in previous_buys if previous_buy.qtty != 0
    ]
    if len(diff_zero) == 0:
        return

    previous_buy = diff_zero[0]
    if sell.price != previous_buy.price or sell.base_cur != previous_buy.base_cur:
        raise CostMethodError(sell, previous_buy.price, previous_buy.base_cur)


def get_lots(txns: List[AdjustedTxn], check: bool) -> List[AdjustedTxn]:
    local_txns = copy.deepcopy(txns)
    checks.check_base_currency(txns)

    buys = [txn for txn in local_txns if txn.qtty >= 0]
    sells = [txn for txn in local_txns if txn.qtty < 0]

    buys_lot: List[AdjustedTxn] = buys if len(sells) == 0 else []
    for sell in sells:
        previous_buys = [txn for txn in buys if txn.date <= sell.date]
        checks.check_short_sell_past(previous_buys, sell)
        later_buys = [txn for txn in buys if txn.date > sell.date]
        sell_qtty = abs(sell.qtty)

        i = 0
        while i < len(previous_buys) and sell_qtty > 0:
            previous_buy = previous_buys[i]
            check_sell(sell, previous_buys, check)
            if sell_qtty >= previous_buy.qtty:
                sell_qtty -= previous_buy.qtty
                previous_buys[i].qtty = 0
            else:
                previous_buys[i].qtty -= sell_qtty
                sell_qtty = 0

            i += 1

        buys_lot = [*previous_buys, *later_buys]

    return buys_lot


def get_sell_lots(
    lots: List[AdjustedTxn], sell_date: str, sell_qtty: float, check: bool
):
    checks.check_short_sell_current(lots, sell_qtty)
    buy_lots = get_lots(lots, check)
    previous_buys = [lot for lot in buy_lots.copy() if lot.date <= sell_date]

    fifo_lots: List[AdjustedTxn] = []
    sell_qtty_curr = sell_qtty

    i = 0
    while sell_qtty_curr > 0 and i < len(lots):
        buy = previous_buys[i]
        if buy.qtty == 0:
            pass
        elif sell_qtty_curr > buy.qtty:
            fifo_lots.append(
                AdjustedTxn(buy.date, buy.price, buy.base_cur, buy.qtty, buy.acct)
            )
            sell_qtty_curr -= buy.qtty
        else:
            fifo_lots.append(
                AdjustedTxn(buy.date, buy.price, buy.base_cur, sell_qtty_curr, buy.acct)
            )
            sell_qtty_curr = 0
        i += 1

    return fifo_lots


def txn2hl(
    txns: List[AdjustedTxn],
    date: str,
    cur: str,
    cash_account: str,
    revenue_account: str,
    value: float,
    sell_cmd: str,
):
    adj_comm = adjust_commodity(cur)
    base_curr = txns[0].base_cur
    avg_cost = get_avg_fifo(txns)
    sum_qtty = sum(txn.qtty for txn in txns)
    price = value / sum_qtty
    dt = datetime.strptime(date, "%Y-%m-%d").date()
    xirr = get_xirr(price, dt, txns) or 0 * 100

    txn_hl = f"""
{date} Sold {cur}  ; cost_method:fifo
    ; commodity:{cur}, qtty:{sum_qtty:,.2f}, price:{price:,.2f}
    ; avg_cost:{avg_cost:,.4f}, xirr:{xirr:.2f}% annual percent rate 30/360US
    ; command:{sell_cmd}
    {cash_account}  {value:.2f} {base_curr}
"""

    for txn in txns:
        txn_hl += f"    {txn.acct}    {txn.qtty * -1} {adj_comm} @ {txn.price} {base_curr}  ; buy_date:{txn.date}, base_cur:{txn.base_cur}\n"

    txn_hl += f"    {revenue_account}   "
    comm = ["hledger", "-f-", "print", "--explicit"]
    txn_proc = subprocess.run(
        comm,
        input=txn_hl.encode(),
        capture_output=True,
    )
    txn_print: str = txn_proc.stdout.decode("utf8")
    return txn_print



from hledger_lots.lib import AdjustedTxn

txns_only_buying = [
    AdjustedTxn(date="2022-01-01", price=10.0, base_cur="USD", qtty=10.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-02", price=20.0, base_cur="USD", qtty=2.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-03", price=30.0, base_cur="USD", qtty=5.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-04", price=15.0, base_cur="USD", qtty=48.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-05", price=25.0, base_cur="USD", qtty=21.0, acct="Acct1"),
]


txns_qtty_never_zero = [
    AdjustedTxn(date="2022-01-01", price=10.0, base_cur="USD", qtty=5.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-02", price=20.0, base_cur="USD", qtty=2.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-03", price=30.0, base_cur="USD", qtty=7.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-04", price=15.0, base_cur="USD", qtty=-3.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-05", price=25.0, base_cur="USD", qtty=2.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-06", price=35.0, base_cur="USD", qtty=5.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-07", price=10.0, base_cur="USD", qtty=-5.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-08", price=20.0, base_cur="USD", qtty=2.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-09", price=30.0, base_cur="USD", qtty=4.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-10", price=15.0, base_cur="USD", qtty=-4.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-11", price=25.0, base_cur="USD", qtty=3.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-12", price=35.0, base_cur="USD", qtty=4.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-13", price=10.0, base_cur="USD", qtty=-5.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-14", price=20.0, base_cur="USD", qtty=3.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-15", price=30.0, base_cur="USD", qtty=2.0, acct="Acct1"),
]


expected_qtty_never_zero = [
    AdjustedTxn(date="2022-01-01", price=10.0, base_cur="USD", qtty=0, acct="Acct1"),
    AdjustedTxn(date="2022-01-02", price=20.0, base_cur="USD", qtty=0, acct="Acct1"),
    AdjustedTxn(date="2022-01-03", price=30.0, base_cur="USD", qtty=0, acct="Acct1"),
    AdjustedTxn(date="2022-01-05", price=25.0, base_cur="USD", qtty=0, acct="Acct1"),
    AdjustedTxn(date="2022-01-06", price=35.0, base_cur="USD", qtty=4.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-08", price=20.0, base_cur="USD", qtty=2.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-09", price=30.0, base_cur="USD", qtty=4.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-11", price=25.0, base_cur="USD", qtty=3.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-12", price=35.0, base_cur="USD", qtty=4.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-14", price=20.0, base_cur="USD", qtty=3.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-15", price=30.0, base_cur="USD", qtty=2.0, acct="Acct1"),
]

expected_qtty_never_zero_sell_some = [
    AdjustedTxn(date='2022-01-06', price=35.0, base_cur='USD', qtty=4.0, acct='Acct1'),
    AdjustedTxn(date='2022-01-08', price=20.0, base_cur='USD', qtty=2.0, acct='Acct1'),
    AdjustedTxn(date='2022-01-09', price=30.0, base_cur='USD', qtty=4.0, acct='Acct1'),
    AdjustedTxn(date='2022-01-11', price=25.0, base_cur='USD', qtty=1.0, acct='Acct1')
]


txns_qtty_reaches_zero = [
    AdjustedTxn(date="2022-01-01", price=10.0, base_cur="USD", qtty=5.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-02", price=20.0, base_cur="USD", qtty=2.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-03", price=30.0, base_cur="USD", qtty=-3.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-04", price=15.0, base_cur="USD", qtty=4.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-05", price=25.0, base_cur="USD", qtty=2.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-06", price=35.0, base_cur="USD", qtty=-1.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-07", price=10.0, base_cur="USD", qtty=-4.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-08", price=20.0, base_cur="USD", qtty=-5.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-09", price=30.0, base_cur="USD", qtty=10.0, acct="Acct1"),
    AdjustedTxn(
        date="2022-01-10", price=15.0, base_cur="USD", qtty=-10.0, acct="Acct1"
    ),
    AdjustedTxn(date="2022-01-11", price=25.0, base_cur="USD", qtty=5.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-12", price=35.0, base_cur="USD", qtty=5.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-13", price=10.0, base_cur="USD", qtty=-5.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-14", price=20.0, base_cur="USD", qtty=3.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-15", price=30.0, base_cur="USD", qtty=-3.0, acct="Acct1"),
]

expected_qtty_reaches_zero = [
    AdjustedTxn(date="2022-01-01", price=10.0, base_cur="USD", qtty=0, acct="Acct1"),
    AdjustedTxn(date="2022-01-02", price=20.0, base_cur="USD", qtty=0, acct="Acct1"),
    AdjustedTxn(date="2022-01-04", price=15.0, base_cur="USD", qtty=0, acct="Acct1"),
    AdjustedTxn(date="2022-01-05", price=25.0, base_cur="USD", qtty=0, acct="Acct1"),
    AdjustedTxn(date="2022-01-09", price=30.0, base_cur="USD", qtty=0, acct="Acct1"),
    AdjustedTxn(date="2022-01-11", price=25.0, base_cur="USD", qtty=0, acct="Acct1"),
    AdjustedTxn(date="2022-01-12", price=35.0, base_cur="USD", qtty=2.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-14", price=20.0, base_cur="USD", qtty=3.0, acct="Acct1"),
]

expected_qtty_reaches_zero_sell_all = [
    AdjustedTxn(date="2022-01-12", price=35.0, base_cur="USD", qtty=2.0, acct="Acct1"),
    AdjustedTxn(date="2022-01-14", price=20.0, base_cur="USD", qtty=3.0, acct="Acct1"),
]



from hledger_lots.lib import AdjustedTxn
import pytest
from hledger_lots import checks

# TODO: Test --check-no-check option
class TestCheckShortSellPast:
    previous_buys = [
        AdjustedTxn(
            date="2022-01-01", price=100.0, base_cur="USD", qtty=50.0, acct="ABC"
        ),
        AdjustedTxn(
            date="2022-01-15", price=110.0, base_cur="USD", qtty=50.0, acct="ABC"
        ),
        AdjustedTxn(
            date="2022-02-01", price=120.0, base_cur="USD", qtty=50.0, acct="ABC"
        ),
    ]

    def test_short_sell_not_allowed(self):
        sell = AdjustedTxn(
            date="2022-03-01", price=100.0, base_cur="USD", qtty=-331, acct="ABC"
        )

        with pytest.raises(ValueError):
            checks.check_short_sell_past(self.previous_buys, sell)

    def test_short_sell_allowed(self):
        sell = AdjustedTxn(
            date="2022-03-01", price=100.0, base_cur="USD", qtty=-100, acct="ABC"
        )
        checks.check_short_sell_past(self.previous_buys, sell)

    def test_empty_previous(self):
        sell = AdjustedTxn(
            date="2022-03-01", price=100.0, base_cur="USD", qtty=-100, acct="ABC"
        )

        with pytest.raises(ValueError):
            checks.check_short_sell_past([], sell)


class TestShortShellCurrent:
    previous_buys = [
        AdjustedTxn(
            date="2022-01-01", price=100.0, base_cur="USD", qtty=50.0, acct="ABC"
        ),
        AdjustedTxn(
            date="2022-01-15", price=110.0, base_cur="USD", qtty=50.0, acct="ABC"
        ),
        AdjustedTxn(
            date="2022-02-01", price=120.0, base_cur="USD", qtty=50.0, acct="ABC"
        ),
    ]

    def test_sell_less(self):
        checks.check_short_sell_current(self.previous_buys, 140)

    def test_sell_equal(self):
        checks.check_short_sell_current(self.previous_buys, 150)

    def test_sell_more(self):
        with pytest.raises(ValueError):
            checks.check_short_sell_current(self.previous_buys, 160)


class TestCheckBaseCurrency:
    def test_same(self):
        txns = [
            AdjustedTxn(
                date="2022-01-01", price=100.0, base_cur="USD", qtty=50.0, acct="ABC"
            ),
            AdjustedTxn(
                date="2022-01-15", price=110.0, base_cur="USD", qtty=50.0, acct="ABC"
            ),
            AdjustedTxn(
                date="2022-02-01", price=120.0, base_cur="USD", qtty=50.0, acct="ABC"
            ),
        ]
        checks.check_base_currency(txns)

    def test_different(self):

        txns = [
            AdjustedTxn(
                date="2022-01-01",
                price=100.0,
                base_cur="USD",
                qtty=50.0,
                acct="ABC",
            ),
            AdjustedTxn(
                date="2022-01-15",
                price=110.0,
                base_cur="BRL",
                qtty=50.0,
                acct="ABC",
            ),
            AdjustedTxn(
                date="2022-02-01",
                price=120.0,
                base_cur="USD",
                qtty=50.0,
                acct="ABC",
            ),
        ]

        with pytest.raises(checks.MultipleBaseCurrencies):
            checks.check_base_currency(txns)



import subprocess
from dataclasses import dataclass
from datetime import date, datetime
from decimal import Decimal
from typing import List, Optional

from hledger_lots.hl import adjust_txn

from . import checks
from .lib import AdjustedTxn, CostMethodError, adjust_commodity, get_xirr


@dataclass
class AvgCost:
    date: str
    total_qtty: float = 0
    total_amount: float = 0
    avg_cost: float = 0


def check_sell(sell: AdjustedTxn, avg_cost: float, check: bool):
    if not check:
        return

    decimals_price = Decimal(str(sell.price)).as_tuple().exponent
    decimals_avg = Decimal(str(avg_cost)).as_tuple().exponent
    if type(decimals_price) == int and type(decimals_avg) == int:
        decimals = min(abs(decimals_price), abs(decimals_avg))
    else:
        raise ValueError("Not a decimal")

    if abs(sell.price - avg_cost) > 10 ** (-decimals):
        raise CostMethodError(sell, avg_cost, sell.base_cur)
    pass


def get_avg_cost(
    txns: List[AdjustedTxn], check: bool, until: Optional[date] = None
) -> List[AvgCost]:
    if until:
        included_txns = [
            txn
            for txn in txns
            if datetime.strptime(txn.date, "%Y-%m-%d").date() <= until
        ]
    else:
        included_txns = txns

    checks.check_base_currency(included_txns)

    total_qtty = 0
    total_amount = 0
    avg_cost = 0

    avg_costs: List[AvgCost] = []

    for txn in included_txns:
        total_qtty += txn.qtty

        if txn.qtty >= 0:
            total_amount += txn.qtty * txn.price
        else:
            check_sell(txn, avg_cost, check)
            total_amount += txn.qtty * avg_cost

        avg_cost = total_amount / total_qtty if total_qtty != 0 else 0
        avg_costs.append(AvgCost(txn.date, total_qtty, total_amount, avg_cost))

    return avg_costs


def avg_sell(
    txns: List[AdjustedTxn],
    date: str,
    qtty: float,
    cur: str,
    cash_account: str,
    revenue_account: str,
    comm_account: str,
    value: float,
    check: bool,
    sell_cmd: str,
):
    adj_comm = adjust_commodity(cur)
    checks.check_short_sell_current(txns, qtty)
    checks.check_base_currency(txns)
    checks.check_available(txns, comm_account, qtty)

    sell_date = datetime.strptime(date, "%Y-%m-%d").date()
    avg_cost = get_avg_cost(txns, check)
    cost = avg_cost[-1].avg_cost

    base_curr = txns[0].base_cur
    price = value / qtty
    xirr = get_xirr(price, sell_date, txns) or 0 * 100

    txn_hl = f"""{date} Sold {cur}  ; cost_method:avg_cost
    ; commodity:{cur}, qtty:{qtty:,.2f}, price:{price:,.2f}
    ; xirr:{xirr:.2f}% annual percent rate 30/360US
    ; command:{sell_cmd}
    {cash_account}    {value:.2f} {base_curr}
    {comm_account}    {qtty * -1} {adj_comm} @ {cost} {base_curr}
    {revenue_account}"""

    comm = ["hledger", "-f-", "print", "--explicit"]
    txn_proc = subprocess.run(comm, input=txn_hl.encode(), capture_output=True)

    txn_print: str = txn_proc.stdout.decode("utf8")
    return txn_print



from hledger_lots import lib
from hledger_lots.lib import AdjustedTxn
import pytest
from datetime import date
import pyxirr


class TestGetAvg:
    def test_empty_list(self):
        assert lib.get_avg_fifo([]) == 0

    def test_single_txn(self):
        txn = AdjustedTxn(
            date="2022-01-01", price=100.0, base_cur="USD", qtty=10.0, acct="1234"
        )
        assert lib.get_avg_fifo([txn]) == 100.0

    def test_multiple_txns(self):
        txns = [
            AdjustedTxn(
                date="2022-01-01", price=100.0, base_cur="USD", qtty=10.0, acct="1234"
            ),
            AdjustedTxn(
                date="2022-01-02", price=150.0, base_cur="USD", qtty=5.0, acct="1234"
            ),
            AdjustedTxn(
                date="2022-01-03", price=200.0, base_cur="USD", qtty=20.0, acct="1234"
            ),
        ]
        assert lib.get_avg_fifo(txns) == pytest.approx(164.28, abs=1e-2)

    def test_zero_quantity(self):
        txns = [
            AdjustedTxn(
                date="2022-01-01", price=100.0, base_cur="USD", qtty=10.0, acct="1234"
            ),
            AdjustedTxn(
                date="2022-01-02", price=150.0, base_cur="USD", qtty=0.0, acct="1234"
            ),
            AdjustedTxn(
                date="2022-01-03", price=200.0, base_cur="USD", qtty=20.0, acct="1234"
            ),
        ]
        assert lib.get_avg_fifo(txns) == pytest.approx(166.67, abs=1e-2)



class TestGetXirr:
    txns = [
            AdjustedTxn("2023-01-23", 100, "USD", -1, "acct"),
            AdjustedTxn("2023-02-23", 100, "USD", -1, "acct"),
        ]
    
    def test_xirr_ok(self):
        assert lib.get_xirr(101,date(2023,3,23),self.txns) == pytest.approx(0.0828, abs=1e-4)

    def test_xirr_only_negatives(self):
        with pytest.raises(pyxirr.InvalidPaymentsError):
            lib.get_xirr(0,date(2023,3,23),self.txns)

    def test_xirr_empty(self):
        lib.get_xirr(0,date(2023,3,23),[]) == None

    def test_xir_negative(self):
        assert lib.get_xirr(99,date(2023,3,23),self.txns) == pytest.approx(-0.0772, abs=1e-4)
        


        
        



from datetime import date, datetime
from pathlib import Path
from typing import Tuple

import pytest

from hledger_lots.fifo_info import FifoInfo
from hledger_lots.avg_info import AvgInfo
import sys



def price_journal(price: str):
    prices = f"""
P 2023-01-05 AAPL 25 USD
P 2023-01-20 AAPL 25 USD
P 2023-02-01 AAPL {price}
"""
    return prices


price_tests = [
    (price_journal("35 USD"), 35),
    (price_journal("35USD"), 35),
    (price_journal("USD 35"), 35),
    (price_journal("USD35"), 35),
]


@pytest.fixture()
def journals(tmp_path: Path):
    hl_txns = """2023-01-05 Buy AAPL
    Asset:Stocks                                   3 AAPL @ 5.2 USD
    Asset:Bank

2023-01-10 Buy BRL
    Asset:FOREX                                  55 BRL @@ 10 USD
    Asset:Bank

2023-01-05 Sell AAPL
    Asset:Bank                                    20 BRL
    Asset:Stocks                                   3 AAPL @ 5.2 USD
    Revenue:Capital Gain
    
"""

    hl_txns += price_journal("6 USD")
    file_path = tmp_path.joinpath("data.journal")
    file_path.touch()
    file_path.write_text(hl_txns)
    file_tup = (str(file_path),)
    return file_tup


@pytest.fixture()
def fifo_info(journals: Tuple[str, ...], monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(sys, "stdin", None)
    return FifoInfo(journals, "AAPL", False)

@pytest.fixture()
def avg_info(journals: Tuple[str, ...], monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(sys, "stdin", None)
    return AvgInfo(journals, "AAPL", False)



class TestInfo:
    @pytest.mark.parametrize("journal,expected", price_tests)
    def test_last_price(
        self,
        tmp_path: Path,
        journal: str,
        expected: float,
        monkeypatch: pytest.MonkeyPatch,
    ):
        file_path = tmp_path.joinpath("data.journal")
        file_path.touch()
        file_path.write_text(journal)
        file_tup = (str(file_path),)

        monkeypatch.setattr(sys, "stdin", None)
        fifo_info = FifoInfo(file_tup, "AAPL", False)

        assert fifo_info.last_price == (date(2023, 2, 1), expected)

    def test_lots_xirr(self, fifo_info: FifoInfo):
        last_buy_date_str = fifo_info.last_buy_date
        if not last_buy_date_str:
            raise ValueError("last_buy_date is None")
        else:
            last_buy_date = datetime.strptime(last_buy_date_str, "%Y-%m-%d").date()

            assert fifo_info.get_lots_xirr(last_buy_date) == pytest.approx(
                6.2528, abs=0.0001
            )


class TestLotInfo:
    expected = {
            "comm": "AAPL",
            "cur": "USD",
            "qtty": "6",
            "amount": "31.20",
            "avg_cost": "5.2000",
            "mkt_price": "6.0000",
            "mkt_amount": "36.00",
            "mkt_profit": "4.80",
            "mkt_date": "2023-02-01",
            "xirr": "6.2529%",
        }
    def test_fifo_info(self, fifo_info: FifoInfo):
        assert fifo_info.info == self.expected

    def test_avg_info(self, avg_info: AvgInfo):
        assert avg_info.info == self.expected

    



import functools
import os
from pathlib import Path
from typing import Literal, Tuple

import rich_click as click

from .avg import avg_sell, get_avg_cost
from .avg_info import AllAvgInfo, AvgInfo
from .fifo import get_lots, get_sell_lots, txn2hl
from .fifo_info import AllFifoInfo, FifoInfo
from .files import get_default_file, get_file_path
from .hl import hledger2txn
from .info import AllInfo, get_commodities
from .lib import default_fn_bool, dt_list2table, get_sell_comm
from .prices_yahoo import get_hledger_prices

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

click.rich_click.USE_MARKDOWN = True
click.rich_click.SHOW_ARGUMENTS = True
click.rich_click.GROUP_ARGUMENTS_OPTIONS = True
click.rich_click.MAX_WIDTH = 80
click.rich_click.SHOW_METAVARS_COLUMN = False
click.rich_click.APPEND_METAVARS_HELP = True
click.rich_click.STYLE_ERRORS_SUGGESTION = "magenta italic"
click.rich_click.ERRORS_SUGGESTION = (
    "Try running the '--help' flag for more information."
)
click.rich_click.ERRORS_EPILOGUE = "To find out more, visit [link=https://github.com/edkedk99/hledger-lots]https://github.com/edkedk99/hledger-lots[/link]"
click.rich_click.STYLE_OPTIONS_TABLE_LEADING = 1
click.rich_click.STYLE_OPTIONS_TABLE_BOX = "SIMPLE"
click.rich_click.STYLE_OPTIONS_PANEL_BORDER = "dim"  # Possibly conceal


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-f",
    "--file",
    type=click.Path(),
    required=False,
    callback=get_file_path,
    multiple=True,
    help="Inform the journal file path. If \"-\", read from stdin. Without this flag read from $LEDGER_FILE or ~/.hledger.journal in this order  or '-f-'.",
)
@click.version_option()
def cli(file: str):  # pyright:ignore
    """
    Commands to apply FIFO(first-in-first-out) or AVERAGE COST accounting principles without manual management of lots. Useful for transactions involving buying and selling foreign currencies or stocks.

    To find out more, visit [https://github.com/edkedk99/hledger-lots](https://github.com/edkedk99/hledger-lots)
    """


@click.command()
@click.option(
    "-f",
    "--file",
    type=click.Path(),
    required=False,
    callback=get_file_path,
    multiple=True,
    help="Inform the journal file path. If \"-\", read from stdin. Without this flag read from $LEDGER_FILE or ~/.hledger.journal in this order  or '-f-'.",
)
@click.option(
    "-g",
    "--avg-cost",
    is_flag=True,
    default=default_fn_bool("HLEDGER_LOTS_AVG_COST", False),
    help='Change cost method to "average cost"". Can be set with env HLEDGER_LOTS_IS_AVG_COST=true|false. Default to false',
)
@click.option(
    "-c",
    "--commodity",
    type=click.STRING,
    prompt=True,
    help="Commodity to get fifo lots",
)
@click.option(
    "-n",
    "--no-desc",
    type=click.STRING,
    default=lambda: os.environ.get("HLEDGER_LOTS_NO_DESC", None),
    prompt=False,
    help="Description to be filtered out from calculation. Needed when closing journal with '--show-costs' option. Works like: not:desc:<value>. Will not be prompted if absent. If closed with default description, the value of this option should be: 'opening|closing balances'. Can be set with env HLEDGER_LOTS_NO_DESC",
)
@click.option(
    "--check/--no-check",
    default=default_fn_bool("HLEDGER_LOTS_CHECK", False),
    help="Enable/Disable check on the commodities previous transactions to ensure it is following the choosen method. Can be set with env HLEDGER_LOTS_CHECK=true|false. Default to false. Inthe future it will default to true",
)
# TODO: fill help
@click.option(
    "-p",
    "--append-prices-to",
    type=click.Path(),
    default=lambda: os.environ.get("HLEDGER_APPEND_PRICES_TO", None),
    prompt=False,
    required=False,
    help="Download market price and append to this option file value. Check the doc for info on how to set it up. Can be set with env HLEDGER_APPEND_PRICES_TO",
)
def view(
    file: Tuple[str, ...],
    avg_cost: bool,
    commodity: str,
    no_desc: str,
    check: bool,
    append_prices_to: Path,
):
    """
    Report lots for a commodity.\r

    Show a report with lots for a commodity considering eventual past sale using FIFO or AVERAGE COST accounting principles.

    Also show some indicators about the lots and performance if there is prices in the journal after the last purchase. See the docs for details

    All flags, except '--file' will be interactively prompted if absent, much like 'hledger-add'.
    """

    journals = file or get_default_file()
    adj_txn = hledger2txn(journals, commodity, no_desc)

    if append_prices_to:
        get_hledger_prices(file, append_prices_to)

    if avg_cost:
        buy_lots = get_avg_cost(adj_txn, check)
        table = dt_list2table(buy_lots)
        click.echo(table)
        avg_info = AvgInfo(journals, commodity, check)
        click.echo(avg_info.info_txt)

    else:
        buy_lots = get_lots(adj_txn, check)
        table = dt_list2table(buy_lots)
        click.echo(table)
        fifo_info = FifoInfo(journals, commodity, check)
        click.echo(fifo_info.info_txt)


@click.command()
@click.option(
    "-f",
    "--file",
    type=click.Path(),
    required=False,
    callback=get_file_path,
    multiple=True,
    help="Inform the journal file path. If \"-\", read from stdin. Without this flag read from $LEDGER_FILE or ~/.hledger.journal in this order  or '-f-'.",
)
@click.option(
    "-g",
    "--avg-cost",
    is_flag=True,
    default=default_fn_bool("HLEDGER_LOTS_AVG_COST", False),
    help='Change cost method to "average cost"". Can be set with env HLEDGER_LOTS_IS_AVG_COST=true|false. Default to false',
)
@click.option(
    "-c",
    "--commodity",
    type=click.STRING,
    prompt=True,
    help="Commodity you are selling",
)
@click.option(
    "-n",
    "--no-desc",
    type=click.STRING,
    default=lambda: os.environ.get("HLEDGER_LOTS_NO_DESC", None),
    prompt=False,
    help="Description to be filtered out from calculation. Needed when closing journal with '--show-costs' option. Works like: not:desc:<value>. Will not be prompted if absent. If closed with default description, the value of this option should be: 'opening|closing balances'. Can be set with env HLEDGER_LOTS_NO_DESC",
)
@click.option(
    "--check/--no-check",
    default=default_fn_bool("HLEDGER_LOTS_CHECK", False),
    help="Enable/Disable check on the commodities previous transactions to ensure it is following the choosen method. Can be set with env HLEDGER_LOTS_CHECK=tru|false. Default to false. Inthe future it will default to true",
)
@click.option(
    "-s",
    "--commodity-account",
    type=click.STRING,
    prompt=False,
    required=False,
    help="What account holds product of the sale. Only used with --avg-cost",
)
@click.option(
    "-a",
    "--cash-account",
    type=click.STRING,
    prompt=True,
    help="What account entered the product of the sell",
)
@click.option(
    "-r",
    "--revenue-account",
    type=click.STRING,
    prompt=True,
    help="Account that represent capital gain/loss",
)
@click.option(
    "-d",
    "--date",
    type=click.STRING,
    prompt=True,
    help="Date of the transaction (YYYY-mm-dd)",
)
@click.option(
    "-q",
    "--quantity",
    help="Quantity being sold",
    type=click.FLOAT,
    prompt=True,
)
@click.option(
    "-p",
    "--price",
    help="Price being sold",
    type=click.FLOAT,
    prompt=True,
)
@click.pass_context
def sell(
    ctx: click.Context,  # pyright:ignore
    file: Tuple[str, ...],
    avg_cost: bool,
    commodity: str,
    no_desc: str,
    commodity_account: str,
    cash_account: str,
    revenue_account: str,
    date: str,
    quantity: float,
    price: float,
    check: bool,
):
    """
    Create a transaction with automatic FIFO or AVERAGE COST for a commodity.\r

    Generate a transaction that represents a sale of a commodity with the following postings:

    - First posting: Positive amount on the 'base-currency' in the account that receives the product of the sale.

    - Multiple lot postings: Each posting is a different lot you are selling for the cost price on purchasing date, calculated according to FIFO accounting principles.

    - Revenue posting: posting that represent Capital Gain or Loss as the difference between the total cost and the amount received on the base-currency.

    This command also add transaction's comment with some indicators. See the docs for more info.

    All flags, except '--file' will be interactively prompted if absent, much like 'hledger-add'.
    """

    journals = file or get_default_file()
    adj_txns = hledger2txn(journals, commodity, no_desc)
    value = quantity * price

    sell_comm_fn = functools.partial(
        get_sell_comm,
        commodity,
        no_desc,
        commodity_account,
        cash_account,
        revenue_account,
        date,
        quantity,
        price,
    )

    if avg_cost and not commodity_account:
        commodity_account = input("Commodity account: ")

    if avg_cost:
        sell_cmd = sell_comm_fn(True)
        sell_avg = avg_sell(
            txns=adj_txns,
            date=date,
            qtty=quantity,
            cur=commodity,
            cash_account=cash_account,
            revenue_account=revenue_account,
            comm_account=commodity_account,
            value=value,
            check=check,
            sell_cmd=sell_cmd,
        )
        click.echo(sell_avg)
    else:
        sell_fifo = get_sell_lots(adj_txns, date, quantity, check)
        sell_cmd = sell_comm_fn(False)
        txn_print = txn2hl(
            txns=sell_fifo,
            date=date,
            cur=commodity,
            cash_account=cash_account,
            revenue_account=revenue_account,
            value=value,
            sell_cmd=sell_cmd,
        )
        click.echo(txn_print)


@click.command(name="list")
@click.option(
    "-f",
    "--file",
    type=click.Path(),
    required=False,
    callback=get_file_path,
    multiple=True,
    help="Inform the journal file path. If \"-\", read from stdin. Without this flag read from $LEDGER_FILE or ~/.hledger.journal in this order  or '-f-'.",
)
@click.option(
    "-g",
    "--avg-cost",
    is_flag=True,
    default=default_fn_bool("HLEDGER_LOTS_AVG_COST", False),
    help='Change cost method to "average cost"". Can be set with env HLEDGER_LOTS_IS_AVG_COST=true|false. Default to false',
)
@click.option(
    "-o",
    "--output-format",
    type=click.Choice(["plain", "pretty", "csv"]),
    default="plain",
    help="Format to output the report",
)
@click.option(
    "-n",
    "--no-desc",
    type=click.STRING,
    default=lambda: os.environ.get("HLEDGER_LOTS_NO_DESC", None),
    prompt=False,
    help="Description to be filtered out from calculation. Needed when closing journal with '--show-costs' option. Works like: not:desc:<value>. Will not be prompted if absent. If closed with default description, the value of this option should be: 'opening|closing balances'. Can be set with env HLEDGER_LOTS_NO_DESC",
)
@click.option(
    "--check/--no-check",
    default=default_fn_bool("HLEDGER_LOTS_CHECK", False),
    help="Enable/Disable check on the commodities previous transactions to ensure it is following the choosen method. Can be set with env HLEDGER_LOTS_CHECK=tru|false. Default to false. Inthe future it will default to true",
)
@click.option(
    "-p",
    "--append-prices-to",
    type=click.Path(),
    default=lambda: os.environ.get("HLEDGER_APPEND_PRICES_TO", None),
    prompt=False,
    required=False,
    help="Download market price and append to this option file value. Check the doc for info on how to set it up. Can be set with env HLEDGER_LOTS_APPEND_PRICES_TO",
)
def list_commodities(
    file: Tuple[str, ...],
    avg_cost: bool,
    output_format: str,
    no_desc: Literal["plain", "pretty", "csv"],
    check: bool,
    append_prices_to: Path,
):
    """
    List indicators for all your commodities in a tabular format sorted from higher to lower **XIRR**. It is advised to use full-screen of the terminal. See the docs for a list of indicators and output examples.

    It can output in three formats: *plain, pretty and csv*.
    """

    journals = file or get_default_file()
    lots_info = AllInfo(journals, no_desc)

    if append_prices_to:
        get_hledger_prices(file, append_prices_to)

    lots_info = (
        AllAvgInfo(file, no_desc, check)
        if avg_cost
        else AllFifoInfo(file, no_desc, check)
    )

    if output_format == "pretty":
        table = lots_info.infos_table("mixed_grid")
    elif output_format == "csv":
        infos_io = lots_info.infos_csv()
        table = infos_io.read()
    else:
        table = lots_info.infos_table("plain")

    click.echo(table)


cli.add_command(view)
cli.add_command(sell)
cli.add_command(list_commodities)



import csv
import re
import subprocess
from dataclasses import dataclass
from datetime import date, datetime
from io import StringIO
from typing import List, Optional, Tuple, TypedDict

from tabulate import tabulate

from .files import get_files_comm
from .hl import hledger2txn
from .lib import adjust_commodity, get_xirr


class LotsInfo(TypedDict):
    comm: str
    cur: str
    qtty: str
    amount: str
    avg_cost: str
    mkt_price: Optional[str]
    mkt_amount: Optional[str]
    mkt_profit: Optional[str]
    mkt_date: Optional[str]
    xirr: Optional[str]


@dataclass
class Price:
    date: date
    comm: str
    price: float
    cur: str


def get_last_price(files_comm: List[str], commodity: str):
    prices_comm = [
        "hledger",
        *files_comm,
        "prices",
        f"cur:{commodity}",
        "--infer-reverse-prices",
    ]
    prices_proc = subprocess.run(prices_comm, capture_output=True)
    prices_str = prices_proc.stdout.decode("utf8")

    if prices_str == "":
        return (None, None)

    prices_list = [row.split(" ", 3) for row in prices_str.split("\n") if row != ""]

    date_list = [
        (row[1], re.sub(r"[a-zA-Z]|\,|\s", "", row[3]))
        for row in prices_list
        if row[2] == adjust_commodity(commodity)
    ]

    if len(date_list) == 0:
        return (None, None)

    last_date_str = date_list[-1][0]
    last_date = datetime.strptime(last_date_str, "%Y-%m-%d").date()
    last_price = float(date_list[-1][1])
    return (last_date, last_price)


def get_commodities(journals: Tuple[str, ...]):
    files_comm = get_files_comm(journals)
    comm = ["hledger", *files_comm, "commodities"]
    commodities_proc = subprocess.run(comm, capture_output=True)
    commodities_str = commodities_proc.stdout.decode("utf8")

    commodities_list = [com for com in commodities_str.split("\n")]
    return commodities_list


class Info:
    def __init__(
        self, journals: Tuple[str, ...], commodity: str, no_desc: Optional[str] = None
    ) -> None:
        self.journals = journals
        self.files_comm = get_files_comm(journals)
        self.commodity = commodity
        self.txns = hledger2txn(journals, commodity, no_desc)

        self.has_txn = len(self.txns) > 0

        self.last_price = get_last_price(self.files_comm, commodity)

        self.market_date, self.market_price = self.last_price

    def get_lots_xirr(self, last_buy_date: date):
        if self.market_date and self.market_price and self.market_date >= last_buy_date:
            xirr = get_xirr(self.market_price, self.market_date, self.txns)
            return xirr

    def get_info_txt(self, info: LotsInfo):
        info_txt = f"""
Info
----
Commodity:      {info['comm']}
Quantity:       {info['qtty']}
Amount:         {info['amount']}
Average Cost:   {info['avg_cost']}
"""

        if self.market_date or self.market_price:
            info_txt += f"""
Market Price:  {info['mkt_price']}
Market Amount: {info['mkt_amount']}
Market Profit: {info['mkt_profit']}
Market Date:   {info['mkt_date']}
Xirr:          {info['xirr']} (APR 30/360US)
"""
        else:
            info_txt += "\nMarket Data not available"

        return info_txt


class AllInfo:
    def __init__(self, journals: Tuple[str, ...], no_desc: str) -> None:
        self.journals = journals
        self.no_desc = no_desc
        self.commodities = get_commodities(journals)

    def get_infos_table(self, infos: List[LotsInfo], output_format: str):
        infos_list = [info for info in infos]
        infos_sorted = sorted(
            infos_list, key=lambda info: info["xirr"] or "", reverse=True
        )
        table = tabulate(
            infos_sorted,
            headers="keys",
            numalign="decimal",
            floatfmt=",.4f",
            tablefmt=output_format,
        )
        return table

    def get_infos_csv(self, infos: List[LotsInfo]):
        infos_list = [info for info in infos]
        infos_sorted = sorted(
            infos_list, key=lambda info: info["xirr"] or "", reverse=True
        )

        fieldnames = infos_sorted[0].keys()
        infos_io = StringIO()
        writer = csv.DictWriter(infos_io, fieldnames=fieldnames)
        writer.writeheader()
        writer.writerows(infos_sorted)
        infos_io.seek(0)
        return infos_io



from datetime import datetime
from typing import Tuple

from .checks import MultipleBaseCurrencies
from .fifo import get_lots
from .hl import hledger2txn
from .info import AllInfo, Info, LotsInfo
from .lib import get_avg_fifo


class FifoInfo(Info):
    def __init__(self, journals: Tuple[str, ...], commodity: str, check: bool):
        super().__init__(journals, commodity)
        self.check = check

        self.lots = get_lots(self.txns, check)
        self.last_buy_date = self.lots[-1].date if len(self.lots) > 0 else None

    @property
    def info(self):
        commodity = self.commodity

        cur = self.lots[0].base_cur
        qtty = sum(lot.qtty for lot in self.lots)
        amount = sum(lot.price * lot.qtty for lot in self.lots)
        avg_cost = get_avg_fifo(self.lots) if qtty > 0 else 0

        if self.has_txn:
            last_buy_date_str = self.lots[-1].date
            last_buy_date = datetime.strptime(last_buy_date_str, "%Y-%m-%d").date()
            xirr = self.get_lots_xirr(last_buy_date)
        else:
            xirr = 0

        if self.market_price and self.market_date and xirr:
            market_price_str = f"{self.market_price:,.4f}"
            market_amount = self.market_price * qtty
            market_amount_str = f"{market_amount:,.2f}"
            market_profit = market_amount - amount
            market_profit_str = f"{market_profit:,.2f}"
            market_date = self.market_date.strftime("%Y-%m-%d")

            xirr_str = f"{xirr:,.4f}%"
        else:
            market_amount_str = ""
            market_profit_str = ""
            market_date = ""
            market_price_str = ""
            xirr_str = ""

        return LotsInfo(
            comm=commodity,
            cur=cur,
            qtty=str(qtty),
            amount=f"{amount:,.2f}",
            avg_cost=f"{avg_cost:,.4f}",
            mkt_price=market_price_str,
            mkt_amount=market_amount_str,
            mkt_profit=market_profit_str,
            mkt_date=market_date,
            xirr=xirr_str,
        )

    @property
    def info_txt(self):
        return self.get_info_txt(self.info)


class AllFifoInfo(AllInfo):
    def __init__(self, journals: Tuple[str, ...], no_desc: str, check: bool):
        super().__init__(journals, no_desc)
        self.check = check

    def get_info(self, commodity: str):
        txns = hledger2txn(self.journals, commodity, self.no_desc)
        try:
            lots = get_lots(txns, self.check)
        except MultipleBaseCurrencies:
            return None

        if len(lots) > 0:
            lot_info = FifoInfo(self.journals, commodity, self.check).info
            return lot_info

    @property
    def infos(self):
        infos = [self.get_info(com) for com in self.commodities]
        infos = [info for info in infos if info is not None]
        return infos

    def infos_table(self, output_format: str):
        return self.get_infos_table(self.infos, output_format)

    def infos_csv(self):
        return self.get_infos_csv(self.infos)



#!/home/eduardo/projects/hledger-lots/venv/bin/python

# $Id: rst2odt.py 8994 2022-01-29 16:28:17Z milde $
# Author: Dave Kuhlman <dkuhlman@rexx.com>
# Copyright: This module has been placed in the public domain.

"""
A front end to the Docutils Publisher, producing OpenOffice documents.
"""

try:
    import locale
    locale.setlocale(locale.LC_ALL, '')
except:
    pass

from docutils.core import publish_cmdline_to_binary, default_description
from docutils.writers.odf_odt import Writer, Reader


description = ('Generates OpenDocument/OpenOffice/ODF documents from '
               'standalone reStructuredText sources.  ' + default_description)


writer = Writer()
reader = Reader()
output = publish_cmdline_to_binary(reader=reader, writer=writer,
                                   description=description)
